import abc
import dataclasses
import typing
from functools import cached_property

import msgspec

from telegrinder.msgspec_utils.encoder import encoder
from telegrinder.tools.keyboard.utils import freaky_keyboard_merge
from telegrinder.tools.serialization.json_ser import JSONSerializer
from telegrinder.tools.serialization.utils import get_model_serializer
from telegrinder.types.objects import (
    CallbackGame,
    CopyTextButton,
    KeyboardButtonPollType,
    KeyboardButtonRequestChat,
    KeyboardButtonRequestUsers,
    LoginUrl,
    SwitchInlineQueryChosenChat,
    WebAppInfo,
)

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

    from telegrinder.bot.rules.abc import ABCRule
    from telegrinder.bot.rules.button import ButtonRule
    from telegrinder.tools.keyboard import keyboard
    from telegrinder.tools.keyboard.base import BaseKeyboard
    from telegrinder.tools.keyboard.button import BaseButton
    from telegrinder.tools.serialization.abc import ABCDataSerializer

type CallbackData = str | dict[str, typing.Any] | DataclassInstance | msgspec.Struct
type Keyboard = keyboard.Keyboard
type InlineKeyboard = keyboard.InlineKeyboard


@dataclasses.dataclass(kw_only=True)
class BaseButton[T: BaseKeyboard = typing.Any](abc.ABC):
    new_row: bool = dataclasses.field(default=False)

    @property
    @abc.abstractmethod
    def keyboard_class(self) -> type[T]:
        pass

    @property
    @abc.abstractmethod
    def rule(self) -> ABCRule:
        pass

    if typing.TYPE_CHECKING:

        def __get__(self, instance: T | None, owner: type[T]) -> ButtonRule[typing.Self]: ...

        @property
        def as_keyboard(self: BaseButton[T]) -> type[T]: ...
    else:

        def as_keyboard(self, *args, **kwargs):
            return self.keyboard_class(*args, **kwargs).add(self)

    def __and__(self, other: object, /) -> T:
        if not isinstance(other, self.keyboard_class | type(self)):
            return NotImplemented
        return freaky_keyboard_merge(self, other)

    def __or__(self, other: object, /) -> T:
        if not isinstance(other, self.keyboard_class | type(self)):
            return NotImplemented
        return freaky_keyboard_merge(self, other, row=True)

    def get_data(self) -> dict[str, typing.Any]:
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None and k != "new_row"}


@dataclasses.dataclass
class Button(BaseButton[Keyboard]):
    text: str
    request_contact: bool = dataclasses.field(default=False, kw_only=True)
    request_location: bool = dataclasses.field(default=False, kw_only=True)
    request_chat: KeyboardButtonRequestChat | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    request_user: KeyboardButtonRequestUsers | None = dataclasses.field(default=None, kw_only=True)
    request_poll: KeyboardButtonPollType | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    web_app: WebAppInfo | None = dataclasses.field(default=None, kw_only=True)

    @cached_property
    def rule(self) -> ABCRule:
        from telegrinder.bot.rules.text import Text

        return Text(self.text)

    @cached_property
    def keyboard_class(self) -> type[Keyboard]:
        from telegrinder.tools.keyboard.keyboard import Keyboard

        return Keyboard


@dataclasses.dataclass
class InlineButton(BaseButton[InlineKeyboard]):
    text: str
    url: str | None = dataclasses.field(default=None, kw_only=True)
    login_url: LoginUrl | None = dataclasses.field(default=None, kw_only=True)
    pay: bool | None = dataclasses.field(default=None, kw_only=True)
    callback_data: CallbackData | None = dataclasses.field(default=None, kw_only=True)
    callback_data_serializer: ABCDataSerializer[typing.Any] | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    callback_game: CallbackGame | None = dataclasses.field(default=None, kw_only=True)
    copy_text: str | CopyTextButton | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_current_chat: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_chosen_chat: SwitchInlineQueryChosenChat | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    web_app: str | WebAppInfo | None = dataclasses.field(default=None, kw_only=True)

    def __post_init__(self) -> None:
        model_serializer = get_model_serializer(self.callback_data)

        self.input_callback_data = self.callback_data
        self.callback_data_serializer = self.callback_data_serializer or (
            None if model_serializer is None else model_serializer(type(self.callback_data))
        )

        if (
            self.callback_data_serializer is None
            and isinstance(self.callback_data, msgspec.Struct | dict)
            or dataclasses.is_dataclass(self.callback_data)
        ):
            self.callback_data_serializer = self.callback_data_serializer or JSONSerializer(
                type(self.callback_data),
            )

        if self.callback_data_serializer is not None:
            self.callback_data = self.callback_data_serializer.serialize(self.callback_data)
        elif self.callback_data is not None and not isinstance(self.callback_data, str):
            self.callback_data = encoder.encode(self.callback_data)

        if isinstance(self.copy_text, str):
            self.copy_text = CopyTextButton(text=self.copy_text)

        if isinstance(self.web_app, str):
            self.web_app = WebAppInfo(url=self.web_app)

    @cached_property
    def rule(self) -> ABCRule:
        from telegrinder.bot.rules.payload import PayloadEqRule, PayloadJsonEqRule, PayloadModelRule

        if isinstance(self.input_callback_data, str):
            return PayloadEqRule(self.input_callback_data)

        if isinstance(self.input_callback_data, dict):
            return PayloadJsonEqRule(self.input_callback_data)

        if self.input_callback_data is not None:
            return PayloadModelRule(
                type(self.input_callback_data),
                payload=self.input_callback_data,
                serializer=type(self.callback_data_serializer) if self.callback_data_serializer else None,
            )

        raise ValueError("Cannot create rule, because callback data is not defined.")

    @cached_property
    def keyboard_class(self) -> type[InlineKeyboard]:
        from telegrinder.tools.keyboard.keyboard import InlineKeyboard

        return InlineKeyboard


__all__ = ("BaseButton", "Button", "InlineButton")



from __future__ import annotations

import re
import typing

from nodnod.scope import Scope
from vbml.patcher.abc import ABCPatcher
from vbml.patcher.patcher import Patcher

from telegrinder.modules import logger
from telegrinder.tools.global_context.global_context import GlobalContext, ctx_var, runtime_init
from telegrinder.tools.loop_wrapper import LoopWrapper


@runtime_init
class TelegrinderContext(GlobalContext, thread_safe=True):
    """The thread-safe type-hinted telegrinder context.

    Example:
    ```
    from telegrinder.tools.global_context import GlobalContext, TelegrinderContext

    telegrinder_ctx = TelegrinderContext(...)  # with type-hints
    assert telegrinder_ctx == GlobalContext("telegrinder")  # ok
    ```

    """

    __ctx_name__ = "telegrinder"

    node_global_scope: typing.Final[Scope] = ctx_var(
        default_factory=lambda: Scope(detail="global"),
        init=False,
        const=True,
    )
    vbml_pattern_flags: re.RegexFlag | None = ctx_var(default=None, init=False)
    vbml_patcher: ABCPatcher = ctx_var(default_factory=Patcher, init=False)
    loop_wrapper: typing.Final[LoopWrapper] = ctx_var(default_factory=LoopWrapper, init=False, const=True)

    async def close_global_scope(self) -> None:
        await self.node_global_scope.close()
        logger.debug("Node global scope closed")


__all__ = ("TelegrinderContext",)



# pyright: reportMissingImports=none, reportAttributeAccessIssue=none, reportMissingModuleSource=none

import asyncio
import contextvars
import inspect
import logging
import os
import pathlib
import re
import shlex
import sys
import traceback
import types
import typing
from annotationlib import type_repr
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler, WatchedFileHandler

import betterconf
from choicelib import choice_in_order
from kungfu.library.monad.option import NOTHING, Option, Some

if typing.TYPE_CHECKING:
    from _typeshed import OptExcInfo
    from loguru import FileHandlerConfig as _LoguruFileHandler
else:

    class _LoguruFileHandler(dict):
        def __new__(cls, *args, **kwargs):
            if logging_module != "loguru":
                raise ModuleNotFoundError("FileHandlerConfig uses for loguru.") from None

            return super().__new__(cls)


type LoggerModule = typing.Literal["logging", "loguru", "structlog"]
type Sink = typing.TextIO | typing.Any
type LoggerLevel = typing.Literal["DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "EXCEPTION"]

_LoggingFileHandler = RotatingFileHandler | TimedRotatingFileHandler | WatchedFileHandler | logging.FileHandler


class Colors:
    RESET = "\033[0m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    BLACK = "\033[30m"
    LIGHT_RED = "\033[91m"
    LIGHT_GREEN = "\033[92m"
    LIGHT_YELLOW = "\033[93m"
    LIGHT_BLUE = "\033[94m"
    LIGHT_MAGENTA = "\033[95m"
    LIGHT_CYAN = "\033[96m"
    LIGHT_WHITE = "\033[97m"
    LIGHT_BLACK = "\033[90m"


LOGGER_MODULE: typing.Final = typing.cast(
    "LoggerModule",
    choice_in_order(["structlog", "loguru"], default="logging", do_import=False),
)
DEFAULT_LOGGING_FORMAT: typing.Final = (
    "<light_white>{name: <4} |</light_white> <level>{levelname: <8}</level>"
    " <light_white>|</light_white> <light_green>{asctime}</light_green> <light_white>"
    "|</light_white> <level_module>{module}</level_module><light_white>:</light_white>"
    "<func_name>{funcName}</func_name><light_white>:</light_white><lineno>{lineno}</lineno>"
    " <light_white>></light_white> <message>{message}</message>"
)
DEFAULT_STRUCTLOG_FORMAT: typing.Final = (
    "[<light_blue>{name}</light_blue>] {location} "
    "[<light_black>{asctime}</light_black>] "
    "<light_white>~</light_white> {message}"
)
DEFAULT_LOGURU_FORMAT: typing.Final = (
    "telegrinder | <level>{level: <8}</level> | "
    "<lg>{time:YYYY-MM-DD HH:mm:ss}</lg> | "
    "<le>{name}</le>:<le>{function}</le>:"
    "<le>{line}</le> > <lw>{message}</lw>"
)

CALL_STACK_CONTEXT: typing.Final = contextvars.ContextVar[tuple[types.FrameType, "OptExcInfo"]]("_call_stack")

COLORS: typing.Final = dict(
    reset=Colors.RESET,
    red=Colors.RED,
    green=Colors.GREEN,
    blue=Colors.BLUE,
    white=Colors.WHITE,
    yellow=Colors.YELLOW,
    magenta=Colors.MAGENTA,
    cyan=Colors.CYAN,
    black=Colors.BLACK,
    light_red=Colors.LIGHT_RED,
    light_green=Colors.LIGHT_GREEN,
    light_blue=Colors.LIGHT_BLUE,
    light_white=Colors.LIGHT_WHITE,
    light_yellow=Colors.LIGHT_YELLOW,
    light_magenta=Colors.LIGHT_MAGENTA,
    light_cyan=Colors.LIGHT_CYAN,
    light_black=Colors.LIGHT_BLACK,
)
LEVEL_FORMAT_SETTINGS: typing.Final = dict(
    DEBUG=dict(
        level="light_blue",
        level_module="blue",
        func_name="blue",
        lineno="light_yellow",
        message="light_blue",
    ),
    INFO=dict(
        level="cyan",
        level_module="light_cyan",
        func_name="light_cyan",
        lineno="light_yellow",
        message="light_green",
    ),
    WARNING=dict(
        level="light_yellow",
        level_module="light_magenta",
        func_name="light_magenta",
        lineno="light_blue",
        message="light_yellow",
    ),
    ERROR=dict(
        level="red",
        level_module="light_yellow",
        func_name="light_yellow",
        lineno="light_blue",
        message="light_red",
    ),
    CRITICAL=dict(
        level="magenta",
        level_module="light_red",
        func_name="light_red",
        lineno="light_yellow",
        message="light_magenta",
    ),
)
VARIABLE_NAME_PATTERN: typing.Final = re.compile(r"[A-Za-z_][A-Za-z_0-9]*")
ASSIGNMENT_OPERATOR: typing.Final = "="
LOGGER_LEVELS: typing.Final = ("DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "EXCEPTION")
LOGGER_MODULES: typing.Final = ("logging", "loguru", "structlog")
_NODEFAULT: typing.Final = typing.cast("typing.Any", object())
_LOAD_ENV_FILE = False
_ENV_FILE_NAME = ".env"
_ENV_FILE_PATH: pathlib.Path | None = None
_ANSI_ESCAPE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def _find_env_file() -> Option[pathlib.Path]:
    caller_frame = sys._getframe()

    while caller_frame:
        if caller_frame.f_back is None:
            break

        caller_frame = caller_frame.f_back

    caller_dir = os.path.dirname(caller_frame.f_code.co_filename)
    start_dir = sys.path[0]

    # Handle empty paths (can happen on Windows in certain launch conditions)
    if not caller_dir or not start_dir:
        search_path = caller_dir or start_dir or "."
    else:
        try:
            search_path = os.path.relpath(caller_dir, start_dir)
        except ValueError:
            search_path = caller_dir

    for root, _, files in os.walk(search_path):
        if _ENV_FILE_NAME in files:
            return Some(pathlib.Path(root) / _ENV_FILE_NAME)

    return NOTHING


@typing.overload
def take_env(var_name: str, /) -> str: ...


@typing.overload
def take_env[T](var_name: str, var_type: type[T], /) -> T: ...


@typing.overload
def take_env[T](var_name: str, var_type: type[T], /, *, default: T) -> T: ...


def take_env[T](
    name: str,
    var_type: type[T] = str,
    /,
    *,
    default: T = _NODEFAULT,
) -> T:
    try:
        value = DOTENV.get(name)
    except betterconf.VariableNotFoundError:
        if default is _NODEFAULT:
            raise
        return default

    if var_type is str:
        return value  # type: ignore

    if var_type not in CASTERS:
        raise NotImplementedError(f"Caster for type `{type_repr(var_type)}` is not implemented.")

    return CASTERS[var_type].cast(value)


class _LoggerLevelCaster(betterconf.AbstractCaster):
    def cast(self, value: str) -> LoggerLevel:
        if value not in LOGGER_LEVELS:
            raise betterconf.ImpossibleToCastError(value, self)
        return value


class _LoggerModuleCaster(betterconf.AbstractCaster):
    def cast(self, value: str) -> LoggerModule:
        if value not in LOGGER_MODULES:
            raise betterconf.ImpossibleToCastError(value, self)
        return value


class _DotenvProvider(betterconf.AbstractProvider):
    __slots__ = ("env_file", "loaded")

    env_file: pathlib.Path | None
    loaded: bool

    def __init__(self) -> None:
        self.env_file = None
        self.loaded = False

    def load(self) -> None:
        if self.loaded:
            return

        self.loaded = True
        self.env_file = _ENV_FILE_PATH if _ENV_FILE_PATH is not None else _find_env_file().unwrap_or_none()

        if self.env_file is None:
            return

        variables: dict[str, str] = {}

        for line in self.env_file.read_text().splitlines():
            match tuple(shlex.shlex(instream=line, posix=True)):
                case (var_name, operator, *tokens) if (
                    tokens and operator == ASSIGNMENT_OPERATOR and VARIABLE_NAME_PATTERN.match(var_name)
                ):
                    variables[var_name] = "".join(tokens).replace(r"\n", "\n").replace(r"\t", "\t")
                case _:
                    continue

        os.environ.update(variables)

    def get(self, name: str) -> str:
        if not self.loaded and _LOAD_ENV_FILE is True:
            self.load()
        return ENV.get(name)


DOTENV: typing.Final = _DotenvProvider()
ENV: typing.Final = betterconf.EnvironmentProvider()
CASTERS: typing.Final[dict[type[typing.Any], betterconf.AbstractCaster]] = {
    LoggerLevel: (to_logger_level := _LoggerLevelCaster()),  # type: ignore
    **betterconf.caster.BUILTIN_CASTERS,
}


@betterconf.betterconf(prefix="TELEGRINDER_LOGGER", provider=DOTENV)
class LoggerConfig:
    MODULE: typing.Literal["logging", "loguru", "structlog"] = betterconf.field(
        default=LOGGER_MODULE,
        caster=_LoggerModuleCaster(),
    )
    LEVEL: LoggerLevel = betterconf.field(
        default="DEBUG",
        caster=to_logger_level,
    )
    FORMAT: str | None = betterconf.constant_field(None)
    COLORIZE: bool = betterconf.field(
        default=True,
        caster=betterconf.caster.to_bool,
    )
    FILE_HANDLER_FORMAT: str | None = betterconf.constant_field(None)
    FILE_HANDLER_COLORIZE: bool = betterconf.field(
        default=False,
        caster=betterconf.caster.to_bool,
    )


def _is_async_logger(logger: typing.Any, /) -> bool:
    return all(hasattr(logger, attr) for attr in AnyAsyncLogger.__protocol_attrs__)


class AnyLogger(typing.Protocol):
    def debug(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def info(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def warning(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def error(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def critical(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    def exception(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...


class AnyAsyncLogger(typing.Protocol):
    async def adebug(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    async def ainfo(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    async def awarning(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    async def aerror(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    async def acritical(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...

    async def aexception(self, __msg: str, *args: typing.Any, **kwargs: typing.Any) -> None: ...


class Logger(AnyLogger, AnyAsyncLogger, typing.Protocol):
    def set_logger(self, __logger: AnyLogger) -> None: ...


class WrapperAsyncLogger:
    def __init__(self, logger: Logger, /) -> None:
        self._logger = logger

    def __getattr__(self, __name: str) -> typing.Any:
        if __name in AnyAsyncLogger.__dict__:
            return lambda *args, **kwargs: self._async_log(
                getattr(self._logger, __name.removeprefix("a")),
                *args,
                **kwargs,
            )

        return super().__getattribute__(__name)

    async def _async_log(
        self,
        method: typing.Callable[..., typing.Any],
        /,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        tok = CALL_STACK_CONTEXT.set((sys._getframe(0).f_back, sys.exc_info()))  # type: ignore
        ctx = contextvars.copy_context()

        try:
            await asyncio.get_running_loop().run_in_executor(
                executor=None,
                func=lambda: ctx.run(lambda: method(*args, **kwargs)),
            )
        finally:
            CALL_STACK_CONTEXT.reset(tok)


class LoggingFormatter(logging.Formatter):
    def __init__(self, format: str, colorize: bool, logger_module: str, /) -> None:
        self.level_formats = _get_level_format(format, colorize)
        self.colorize = colorize
        self.logger_module = logger_module
        super().__init__()

    def format(self, record: logging.LogRecord) -> str:
        if self.logger_module == "logging":
            record = _rich_log_record(record, self.logger_module)

        message = logging.Formatter(
            fmt=self.level_formats.get(record.levelname),
            datefmt="%Y-%m-%d %H:%M:%S",
            style="{",
        ).format(record)

        if not self.colorize:
            message = _remove_ansi_colors(message)

        return message


class LogMessage:
    def __init__(self, fmt: str, args: typing.Any, kwargs: typing.Any) -> None:
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return self.fmt.format(*self.args, **self.kwargs)


class LoggingStyleAdapter(logging.LoggerAdapter):
    logger: logging.Logger

    def __init__(
        self,
        logger: logging.Logger,
        **extra: typing.Any,
    ) -> None:
        super().__init__(logger, extra=extra or None)
        self.log_arg_names = frozenset(inspect.getfullargspec(self.logger._log).args[1:])

    def log(self, level: int, msg: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> None:
        if self.isEnabledFor(level):
            msg, args, kwargs = self.proc(msg, args, kwargs)
            self.logger._log(level, msg, args, **kwargs)

    def proc(
        self,
        msg: typing.Any,
        args: tuple[typing.Any, ...],
        kwargs: dict[str, typing.Any],
    ) -> tuple[typing.Any, tuple[typing.Any, ...], dict[str, typing.Any]]:
        kwargs.setdefault("stacklevel", 2)

        if isinstance(msg, str):
            msg = LogMessage(msg, args, kwargs)
            args = tuple()

        return msg, args, {name: kwargs[name] for name in self.log_arg_names if name in kwargs}


def _remove_handlers(logger: typing.Any, /) -> None:
    for hdlr in logger.handlers[:]:
        logger.removeHandler(hdlr)


def _get_level_format(format: str, colorize: bool, /) -> dict[str, str]:
    level_formats = {}

    for level, settings in LEVEL_FORMAT_SETTINGS.items():
        fmt = format

        for name, color in COLORS.items():
            fmt = fmt.replace(f"<{name}>", color if colorize else "").replace(
                f"</{name}>", COLORS["reset"] if colorize else ""
            )

        for name, color in settings.items():
            fmt = fmt.replace(f"<{name}>", COLORS[color] if colorize else "").replace(
                f"</{name}>", COLORS["reset"] if colorize else ""
            )

        level_formats[level] = fmt

    return level_formats


def _rich_log_record(record: logging.LogRecord, logger_module: str, /) -> logging.LogRecord:
    call_stack, exc_info = CALL_STACK_CONTEXT.get((None, None))
    frame = call_stack or sys._getframe(1)

    if call_stack is None:
        while frame:
            if frame.f_code.co_filename == record.pathname and frame.f_lineno == record.lineno:
                if logger_module == "structlog":
                    frame = frame.f_back.f_back.f_back  # pyright: ignore[reportOptionalMemberAccess]

                break

            frame = frame.f_back

    if record.levelno >= logging.ERROR and record.exc_info is not None:
        if exc_info is not None and any(exc_info):
            record.exc_info = exc_info
        elif logger_module == "structlog":
            record.exc_info = None

    if frame is not None:
        record.funcName = frame.f_code.co_name
        record.module = frame.f_globals.get("__name__", "<module>")
        record.lineno = frame.f_lineno

    if not record.funcName or record.funcName == "<module>":
        record.funcName = "\b"

    return record


def _loguru_filter(record: dict[str, typing.Any]) -> bool:
    if record["extra"].get("telegrinder", False) is not True:
        return False

    frame, exc_info = CALL_STACK_CONTEXT.get((None, None))

    if frame is not None:
        if "file" in record:
            file_name = frame.f_code.co_filename
            record["file"].name = os.path.basename(file_name)
            record["file"].path = file_name

        record["name"] = frame.f_globals.get("__name__", "<module>")
        record["module"] = record["name"].split(".")[-1]
        record["line"] = frame.f_lineno
        record["function"] = frame.f_code.co_name

    if (
        exc_info is not None
        and any(exc_info)
        and record["exception"] is not None
        and not any(record["exception"])
        and record["level"].no >= logger.logger.level("ERROR").no
    ):
        from loguru._recattrs import RecordException  # type: ignore

        record["exception"] = RecordException(exc_info[0], exc_info[1], exc_info[2])  # type: ignore

    return True


def _remove_ansi_colors(text: str) -> str:
    return _ANSI_ESCAPE.sub("", text)


class _json:  # noqa: N801
    def __getattr__(self, __name: str) -> typing.Any:
        from telegrinder.msgspec_utils import json

        return getattr(json, __name)

    def __repr__(self) -> str:
        return "<module 'telegrinder.msgspec_utils.json'>"


class _LoggerProxy:
    def __init__(self) -> None:
        self.logger = None
        self.logger_module = None

    def __repr__(self) -> str:
        return "<LoggerProxy {}: {}>".format(
            self.logger_module or "unknown",
            "(NOT SETUP)" if self.logger is None else repr(self.logger),
        )

    def __await__(self) -> typing.Generator[typing.Any, typing.Any, typing.Any]:
        return iter(())  # type: ignore

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Self:
        return self

    def __getattr__(self, __name: str) -> typing.Any:
        if self.logger is None:
            if __name in ("exception", "aexception") and (exc := sys.exception()) is not None:
                traceback.print_exception(exc, chain=True, file=sys.stderr)

            return self

        if __name in AnyLogger.__dict__ or __name in AnyAsyncLogger.__dict__:
            is_async = __name.startswith("a")

            if self.logging_module is not None:
                level_name = __name.removeprefix("a").upper()
                level_name = (
                    "INFO"
                    if level_name == "SUCCESS" and self.logger_module != "loguru"
                    else "ERROR"
                    if level_name == "EXCEPTION"
                    else level_name
                )
                level = (
                    logging._nameToLevel.get(level_name, logging.NOTSET)
                    if self.logger_module != "loguru"
                    else level_name
                )
                if not hasattr(self.logger, "isEnabledFor") and self.logger.isEnabledFor(level):  # type: ignore
                    return self

            return getattr(self.logger if not is_async else self.async_logger, __name)

        return self

    def set_logger(self, logger: Logger, logger_module: LoggerModule | None = None) -> None:
        self.logger = logger if not isinstance(logger, logging.Logger) else LoggingStyleAdapter(logger)
        self.async_logger = WrapperAsyncLogger(logger) if not _is_async_logger(logger) else logger
        self.logger_module = logger_module


class _SetupLoggerKwargs(typing.TypedDict):
    module: typing.NotRequired[LoggerModule]
    level: typing.NotRequired[LoggerLevel]
    format: typing.NotRequired[str]
    colorize: typing.NotRequired[bool]
    console_sink: typing.NotRequired[Sink | None]
    file: typing.NotRequired[FileHandlerConfig]


class _LoguruFileHandlerConfig(_LoguruFileHandler):
    pass


class FileHandlerConfig:
    handler: _LoggingFileHandler | _LoguruFileHandlerConfig
    format: str
    colorize: bool

    def __init__(
        self,
        handler: _LoggingFileHandler | _LoguruFileHandlerConfig,
        format: str = _NODEFAULT,
        colorize: bool = _NODEFAULT,
        logger_module: str = LOGGER_MODULE,
    ) -> None:
        config = LoggerConfig()
        format = (config.FILE_HANDLER_FORMAT if format is _NODEFAULT else format) or (
            DEFAULT_LOGURU_FORMAT
            if logger_module == "loguru"
            else DEFAULT_STRUCTLOG_FORMAT
            if logger_module == "structlog"
            else DEFAULT_LOGGING_FORMAT
        )
        colorize = config.FILE_HANDLER_COLORIZE if colorize is _NODEFAULT else colorize

        self.handler = handler

        if isinstance(self.handler, dict):
            self.format = self.handler.setdefault("format", format)
            self.colorize = self.handler.setdefault("colorize", colorize)
        else:
            self.format = format
            self.colorize = colorize

            if handler.formatter is None:
                handler.setFormatter(LoggingFormatter(format, colorize, logger_module))

    @classmethod
    def from_logging(
        cls,
        handler: _LoggingFileHandler,
        format: str = _NODEFAULT,
        colorize: bool = _NODEFAULT,
    ) -> typing.Self:
        return cls(handler, format, colorize)

    @classmethod
    def from_loguru(
        cls,
        **kwargs: typing.Unpack[_LoguruFileHandlerConfig],  # type: ignore
    ) -> typing.Self:
        return cls(kwargs)  # type: ignore


logger: Logger = typing.cast("Logger", _LoggerProxy())


def _configure_structlog(
    level: LoggerLevel,
    format: str | None,
    colorize: bool,
    console_sink: Sink | None = None,
    file: FileHandlerConfig | None = None,
    /,
) -> None:
    import re
    from contextlib import suppress

    import structlog

    levels_colors = dict(
        debug=COLORS["light_blue"],
        info=COLORS["light_green"],
        warning=COLORS["light_yellow"],
        error=COLORS["light_red"],
        critical=COLORS["light_red"],
    )

    class SLF4JStyleFormatter:
        BRACE_PATTERN = re.compile(r"\{(?:([^}]*?)(?:!([sra]))?)?\}")
        PERCENT_PATTERN = re.compile(r"%(?:\(([^)]+)\))?([sdfrx])")

        def __init__(self, *, remove_positional_args: bool = True, colors: bool = True) -> None:
            self.remove_positional_args = remove_positional_args
            self.colors = colors

        def __call__(
            self,
            logger: typing.Any,
            method_name: str,
            event_dict: typing.MutableMapping[str, typing.Any],
        ) -> typing.MutableMapping[str, typing.Any]:
            args = event_dict.get("positional_args", ())
            event = event_dict.pop("event", "")
            if not isinstance(event, str):
                return event_dict

            log_level = event_dict.get("level", "debug")
            system_fields = {"level", "logger", "timestamp", "positional_args"}
            kwargs = {k: v for k, v in event_dict.items() if k not in system_fields}
            used_kwargs = set()

            with suppress(TypeError, ValueError, IndexError, KeyError):
                if self.BRACE_PATTERN.search(event):
                    event_dict["event"], used_kwargs = self._format_braces(event, args, kwargs, log_level)
                elif self.PERCENT_PATTERN.search(event):
                    event_dict["event"], used_kwargs = self._format_percent(event, args, kwargs, log_level)
                elif args:
                    event_dict["event"] = self._highlight_values(event, args, log_level)
                else:
                    event_dict["event"] = event

            if self.remove_positional_args:
                if "positional_args" in event_dict:
                    del event_dict["positional_args"]
                for key in used_kwargs:
                    if key in event_dict:
                        del event_dict[key]

            return event_dict

        def _colorize(self, value: typing.Any, log_level: str) -> str:
            return f"{levels_colors[log_level]}{value}{Colors.RESET}" if self.colors else value

        def _format_braces(
            self,
            message: str,
            args: tuple[typing.Any, ...],
            kwargs: dict[str, typing.Any],
            log_level: str,
        ) -> tuple[str, set[str]]:
            result = []
            last_end = 0
            arg_index = 0
            used_kwargs = set()

            for match in self.BRACE_PATTERN.finditer(message):
                result.append(message[last_end : match.start()])

                field_name = match.group(1)
                conversion = match.group(2)

                if field_name:
                    if field_name in kwargs:
                        value = kwargs[field_name]
                        used_kwargs.add(field_name)
                    else:
                        result.append(match.group(0))
                        last_end = match.end()
                        continue
                elif arg_index < len(args):
                    value = args[arg_index]
                    arg_index += 1
                else:
                    result.append(match.group(0))
                    last_end = match.end()
                    continue

                if conversion == "r":
                    formatted_value = repr(value)
                elif conversion == "s":
                    formatted_value = str(value)
                elif conversion == "a":
                    formatted_value = ascii(value)
                else:
                    formatted_value = str(value)

                result.append(self._colorize(formatted_value, log_level))
                last_end = match.end()

            result.append(message[last_end:])
            return "".join(result), used_kwargs

        def _format_percent(
            self,
            message: str,
            args: tuple[typing.Any, ...],
            kwargs: dict[str, typing.Any],
            log_level: str,
        ) -> tuple[str, set[str]]:
            used_kwargs = set[str]()

            try:
                has_named = bool(re.search(r"%\([^)]+\)", message))
                has_positional = bool(re.search(r"%[sdfrx]", message))

                if has_named and not has_positional:
                    formatted = message % kwargs
                    used_kwargs = set(kwargs.keys())
                    for value in kwargs.values():
                        formatted = self._highlight_single_value(formatted, value, log_level)
                elif has_positional and not has_named:
                    formatted = message % args
                    for value in args:
                        formatted = self._highlight_single_value(formatted, value, log_level)
                elif has_named and has_positional:
                    temp_formatted = message

                    for key, value in kwargs.items():
                        placeholder = f"%({key})s"
                        if placeholder in temp_formatted:
                            used_kwargs.add(key)
                            replacement = self._colorize(str(value), log_level)
                            temp_formatted = temp_formatted.replace(placeholder, replacement)

                    if args and "%s" in temp_formatted:
                        temp_formatted = temp_formatted % args
                        for value in args:
                            temp_formatted = self._highlight_single_value(temp_formatted, value, log_level)

                    formatted = temp_formatted
                else:
                    formatted = message

                return formatted, used_kwargs
            except (TypeError, KeyError, ValueError):
                if kwargs:
                    try:
                        formatted = message % kwargs
                        used_kwargs = set(kwargs.keys())
                        for value in kwargs.values():
                            formatted = self._highlight_single_value(formatted, value, log_level)
                        return formatted, used_kwargs
                    except (TypeError, KeyError):
                        pass

                if args:
                    try:
                        formatted = message % args
                        for value in args:
                            formatted = self._highlight_single_value(formatted, value, log_level)
                        return formatted, used_kwargs
                    except (TypeError, ValueError):
                        pass

                return message, used_kwargs

        def _highlight_single_value(
            self,
            message: str,
            value: typing.Any,
            log_level: str,
        ) -> str:
            with suppress(Exception):
                for raw in (str(value), repr(value)):
                    if raw in message:
                        pattern = re.compile(rf"(?<!\w){re.escape(raw)}(?!\w)")
                        message = pattern.sub(lambda m: self._colorize(m.group(0), log_level), message, count=1)
                        break

            return message

        def _highlight_values(
            self,
            full_message: str,
            values: typing.Iterable[typing.Any],
            log_level: str,
        ) -> str:
            for value in values:
                full_message = self._highlight_single_value(full_message, value, log_level)
            return full_message

    class LogLevelColumnFormatter:
        def __init__(self, colorize: bool) -> None:
            self.colorize = colorize

        def __call__(self, key: str, value: typing.Any) -> str:
            if self.colorize:
                color = levels_colors[value]
                return f"[{color}{value:^12}{Colors.RESET}]"

            return f"[{value:^12}]"

    class Filter(logging.Filter):
        def __init__(self, colorize: bool) -> None:
            self.colorize = colorize
            super().__init__()

        def filter(self, record: logging.LogRecord) -> bool:
            record = _rich_log_record(record, "structlog")

            if self.colorize:
                level_color = levels_colors[record.levelname.lower()]
                location = (
                    f"{Colors.LIGHT_CYAN}{record.module}{Colors.RESET}:"
                    f"{level_color}{record.funcName}{Colors.RESET}:"
                    f"{Colors.LIGHT_MAGENTA}{record.lineno}{Colors.RESET} "
                )
            else:
                location = f"{record.module}:{record.funcName}:{record.lineno} "

            record.location = location
            return True

    console_renderer = structlog.dev.ConsoleRenderer(colors=True)

    for column in console_renderer._columns:
        if column.key == "level":
            column.formatter = LogLevelColumnFormatter(colorize)
            break

    telegrinder_logger = logging.getLogger("telegrinder")
    telegrinder_logger.setLevel(level)
    _remove_handlers(telegrinder_logger)

    if console_sink is not None:
        console_handler = logging.StreamHandler(console_sink)
        console_handler.setFormatter(LoggingFormatter(format or DEFAULT_STRUCTLOG_FORMAT, colorize, "structlog"))
        console_handler.addFilter(Filter(colorize))
        telegrinder_logger.addHandler(console_handler)

    if file is not None and isinstance(file.handler, _LoggingFileHandler):
        file.handler.addFilter(Filter(file.colorize))
        telegrinder_logger.addHandler(file.handler)

    struct_logger = structlog.wrap_logger(
        logger=telegrinder_logger,
        processors=[
            structlog.stdlib.filter_by_level,
            structlog.stdlib.add_log_level,
            SLF4JStyleFormatter(),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            console_renderer,
        ],
        wrapper_class=structlog.stdlib.BoundLogger,
        context_class=dict,
        cache_logger_on_first_use=True,
    )
    logger.set_logger(struct_logger, "structlog")  # type: ignore


def _configure_loguru(
    level: LoggerLevel,
    format: str | None,
    colorize: bool,
    console_sink: Sink | None = None,
    file: FileHandlerConfig | None = None,
    /,
) -> None:
    import atexit

    from loguru._logger import Core, Logger

    loguru_logger = Logger(
        core=Core(),
        exception=None,
        depth=0,
        record=False,
        lazy=False,
        colors=False,
        raw=False,
        capture=True,
        patchers=[],
        extra=dict(telegrinder=True),
    )

    def is_enabled_for(lvl: str) -> bool:
        try:
            lno = loguru_logger.level(lvl).no
        except ValueError:
            return False

        return any(lno >= x.levelno for x in loguru_logger._core.handlers.values())

    loguru_logger.isEnabledFor = is_enabled_for
    handlers = []

    if console_sink is not None:
        handlers.append(
            dict(
                sink=console_sink,
                level=level,
                enqueue=True,
                colorize=colorize,
                format=format or DEFAULT_LOGURU_FORMAT,
                filter=_loguru_filter,
            ),
        )

    if file is not None and isinstance(file.handler, dict):  # type: ignore
        handlers.append(file.handler)

    if handlers:
        handlers_ids = loguru_logger.configure(handlers=handlers)

        @atexit.register
        def _() -> None:
            for handler_id in handlers_ids:
                loguru_logger.remove(handler_id)

    logger.set_logger(loguru_logger, "loguru")  # type: ignore


def _configure_logging(
    level: LoggerLevel,
    format: str | None,
    colorize: bool,
    console_sink: Sink | None = None,
    file: FileHandlerConfig | None = None,
    /,
) -> None:
    _logger = logging.getLogger("telegrinder")
    _logger.setLevel(level)
    _remove_handlers(_logger)

    if console_sink is not None:
        console_handler = logging.StreamHandler(console_sink)
        console_handler.setFormatter(LoggingFormatter(format or DEFAULT_LOGGING_FORMAT, colorize, "logging"))
        _logger.addHandler(console_handler)

    if file is not None and isinstance(file.handler, _LoggingFileHandler):
        _logger.addHandler(file.handler)

    logger.set_logger(LoggingStyleAdapter(logger=_logger), "logging")  # type: ignore


def setup_logger(**setup_kwargs: typing.Unpack[_SetupLoggerKwargs]) -> Logger:
    if logger.logger is not None:
        return logger

    config = LoggerConfig()
    args = (
        setup_kwargs.get("level", config.LEVEL),
        setup_kwargs.get("format", config.FORMAT),
        setup_kwargs.get("colorize", config.COLORIZE),
        setup_kwargs.get("console_sink", sys.stderr),
        setup_kwargs.get("file"),
    )

    match setup_kwargs.get("module", config.MODULE):
        case "logging":
            _configure_logging(*args)
        case "loguru":
            _configure_loguru(*args)
        case "structlog":
            _configure_structlog(*args)
        case _ as value:
            typing.assert_never(value)

    return logger


def configure_dotenv(
    *,
    load_file: bool = False,
    file_name: str | None = None,
    file_path: str | pathlib.Path | None = None,
) -> None:
    global _LOAD_ENV_FILE, _ENV_FILE_NAME, _ENV_FILE_PATH

    _LOAD_ENV_FILE = load_file

    if file_name and file_path:
        _ENV_FILE_PATH = pathlib.Path(file_path) / file_name
        _ENV_FILE_NAME = file_name

    elif file_name is not None:
        _ENV_FILE_NAME = file_name

    elif file_path is not None:
        _ENV_FILE_PATH = pathlib.Path(file_path)

    if _ENV_FILE_PATH is not None and not _ENV_FILE_PATH.exists():
        raise FileNotFoundError(f"Env file '{_ENV_FILE_PATH!s}' not found")


if typing.TYPE_CHECKING:
    from telegrinder.msgspec_utils import json
else:
    json = _json()


__all__ = (
    "FileHandlerConfig",
    "LoggingStyleAdapter",
    "configure_dotenv",
    "json",
    "logger",
    "setup_logger",
    "take_env",
)



import datetime
import typing

from telegrinder.api.api import API
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.dispatch.waiter_machine.short_state import ShortStateContext
from telegrinder.modules import logger
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.waiter_machine.hasher import Hasher
    from telegrinder.bot.dispatch.waiter_machine.machine import WaiterMachine
    from telegrinder.bot.dispatch.waiter_machine.short_state import ShortState


class WaiterMiddleware(ABCMiddleware):
    def __init__(
        self,
        machine: WaiterMachine,
        hasher: Hasher[typing.Any, typing.Any],
    ) -> None:
        self.machine = machine
        self.hasher = hasher

    async def pre(self, update_cute: UpdateCute, raw_update: Update, api: API, ctx: Context) -> bool:
        if self.hasher not in self.machine.storage:
            return True

        event = update_cute.incoming_update
        key = self.hasher.get_hash_from_data_from_event(event)
        if not key:
            await logger.ainfo("Unable to get hash from event with hasher {!r}", self.hasher)
            return True

        short_state: ShortState | None = self.machine.storage[self.hasher].get(key.unwrap())
        if not short_state:
            return True

        # Just ignore update if state expired, so it will be dropped automatically by WaiterMachine
        if short_state.expiration_date is not None and datetime.datetime.now() >= short_state.expiration_date:
            return True

        preset_context = Context(short_state=short_state)
        if short_state.context is not None:
            preset_context |= short_state.context.context

        if short_state.filter is not None and not await check_rule(short_state.filter, preset_context):
            await logger.adebug("Filter rule {!r} failed!", short_state.filter)
            return True

        handler = FuncHandler(
            function=self.pass_runtime,
            rules=(short_state.release,) if short_state.release is not None else None,
            preset_context=preset_context,
        )

        if not await handler.run(api, raw_update, ctx) and (on_miss := short_state.actions.get("on_miss")) is not None:
            await on_miss.run(api, raw_update, ctx)

        return False

    async def pass_runtime(
        self,
        event: UpdateCute,
        ctx: Context,
        short_state: ShortState,
    ) -> None:
        ctx.initiator = self.hasher
        short_state.context = ShortStateContext(event.incoming_update, ctx)
        short_state.event.set()


__all__ = ("WaiterMiddleware",)



def compute_number(
    default: int | float,
    input_value: int | float,
    conditional_value: int | float,
    /,
) -> int | float:
    return max(default, input_value) * (input_value <= conditional_value) + input_value * (
        input_value >= conditional_value
    )


__all__ = ("compute_number",)



import dataclasses
import typing

from kungfu.library.monad.result import Error as Err
from kungfu.library.monad.result import Ok

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.router.abc import ABCRouter
from telegrinder.bot.dispatch.view.box import ViewBox
from telegrinder.modules import logger
from telegrinder.tools.magic.inspect import get_frame_module_name
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.view.base import ErrorView, EventView, RawEventView, View
    from telegrinder.bot.dispatch.view.media_group import MediaGroupView


@dataclasses.dataclass(kw_only=True)
class Router[
    MessageView: EventView = EventView,
    EditedMessageView: EventView = EventView,
    ChannelPostView: EventView = EventView,
    EditedChannelPostView: EventView = EventView,
    BusinessConnectionView: EventView = EventView,
    BusinessMessageView: EventView = EventView,
    EditedBusinessMessageView: EventView = EventView,
    DeletedBusinessMessagesView: EventView = EventView,
    MessageReactionView: EventView = EventView,
    MessageReactionCountView: EventView = EventView,
    InlineQueryView: EventView = EventView,
    ChosenInlineResultView: EventView = EventView,
    CallbackQueryView: EventView = EventView,
    ShippingQueryView: EventView = EventView,
    PreCheckoutQueryView: EventView = EventView,
    PurchasedPaidMediaView: EventView = EventView,
    PollView: EventView = EventView,
    PollAnswerView: EventView = EventView,
    MyChatMemberView: EventView = EventView,
    ChatMemberView: EventView = EventView,
    ChatJoinRequestView: EventView = EventView,
    ChatBoostView: EventView = EventView,
    RemovedChatBoostView: EventView = EventView,
    MediaGroup: View = MediaGroupView,
    Error: ErrorView = ErrorView,
    RawEvent: RawEventView = RawEventView,
](
    ABCRouter,
    ViewBox[
        MessageView,
        EditedMessageView,
        ChannelPostView,
        EditedChannelPostView,
        BusinessConnectionView,
        BusinessMessageView,
        EditedBusinessMessageView,
        DeletedBusinessMessagesView,
        MessageReactionView,
        MessageReactionCountView,
        InlineQueryView,
        ChosenInlineResultView,
        CallbackQueryView,
        ShippingQueryView,
        PreCheckoutQueryView,
        PurchasedPaidMediaView,
        PollView,
        PollAnswerView,
        MyChatMemberView,
        ChatMemberView,
        ChatJoinRequestView,
        ChatBoostView,
        RemovedChatBoostView,
        MediaGroup,
        Error,
        RawEvent,
    ],
):
    def __post_init__(self) -> None:
        self.name = ":".join((get_frame_module_name(), self.__class__.__name__, hex(id(self))))

    def __repr__(self) -> str:
        return f"<{self.name}>"

    def __hash__(self) -> int:
        return hash(self.name)

    def __bool__(self) -> bool:
        return any(self.views.values()) or bool(self.raw) or bool(self.event_error)

    async def route_view(self, view: View, api: API, update: Update, context: Context) -> bool:
        # Check if the view is applicable to the update

        await logger.adebug(
            "Checking view `{!r}` from router `{!r}` for update (id={}, type={!r})...",
            view,
            self,
            update.update_id,
            update.update_type,
        )

        match await view.check(api, update, context):
            case Ok(_):
                await logger.adebug(
                    "View `{!r}` from router `{!r}` for update (id={}, type={!r}) is happy, processing...",
                    view,
                    self,
                    update.update_id,
                    update.update_type,
                )
                result = await view.process(api, update, context)

                match result:
                    case Err(error) if isinstance(error, Exception):
                        raise error from None

                await logger.ainfo(
                    "Update(id={}, type={!r}) processed with view `{!r}` from router `{!r}`. {}",
                    update.update_id,
                    update.update_type,
                    view,
                    self,
                    result.error if not result else result.value,
                )
                return bool(result)
            case Err(error):
                await logger.adebug(
                    "Checking view `{!r}` from router `{!r}` for update (id={}, type={!r}) failed: {}",
                    view,
                    self,
                    update.update_id,
                    update.update_type,
                    error,
                )

        return False

    async def route(self, api: API, update: Update, context: Context) -> bool:
        try:
            # Filtering non-empty views
            for view in filter(None, self.views.values()):
                # Route the non-empty view
                if await self.route_view(view, api, update, context):
                    return True

            return False
        except Exception as exception:
            if self.event_error:
                context.exceptions_update[self] = exception  # type: ignore

            raise


__all__ = ("Router",)



import abc
import typing

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.abc import Button, RawKeyboard
    from telegrinder.tools.keyboard.button import BaseButton

from telegrinder.tools.keyboard.abc import ABCKeyboard
from telegrinder.tools.keyboard.utils import RowButtons, copy_keyboard, get_keyboard_button_rules, is_dunder

BUTTON_CLASS_KEY: typing.Final = "__button_class__"


class KeyboardMeta(type):
    if not typing.TYPE_CHECKING:

        def __getattribute__(cls, __name: str) -> typing.Any:
            if (
                not is_dunder(__name)
                and ABCKeyboard not in type.__getattribute__(cls, "__bases__")
                and hasattr(cls, BUTTON_CLASS_KEY)
                and (button_rule := get_keyboard_button_rules(cls).get(__name)) is not None
            ):
                return button_rule

            return super().__getattribute__(__name)


class ABCBaseKeyboard(typing._ProtocolMeta, KeyboardMeta):  # type: ignore
    pass


class BaseKeyboard[KeyboardButton: BaseButton = typing.Any](typing.Protocol, metaclass=ABCBaseKeyboard):
    keyboard: RawKeyboard

    __keyboard_instance__: typing.ClassVar[typing.Self]
    __button_class__: typing.ClassVar[type[BaseButton[typing.Self]]]

    @abc.abstractmethod
    def copy(self, **with_changes: typing.Any) -> typing.Self:
        pass

    @abc.abstractmethod
    def __init_subclass__(cls) -> None:
        pass

    def __and__(self, other: object, /) -> typing.Self:
        if not isinstance(other, self.__button_class__ | type(self)):
            return NotImplemented
        return self.add(other) if isinstance(other, self.__button_class__) else self.merge(other)  # type: ignore

    def __or__(self, other: object, /) -> typing.Self:
        if not isinstance(other, self.__button_class__ | type(self)):
            return NotImplemented
        kb = self.row()
        return kb.add(other) if isinstance(other, self.__button_class__) else kb.merge_to_last_row(other)  # type: ignore

    def add(self, button: Button, /) -> typing.Self:
        if not self.keyboard:
            self.row()

        if isinstance(button, RowButtons):
            self.keyboard[-1].extend(button.get_data())
            if button.auto_row:
                self.row()
            return self

        if isinstance(button, list):
            self.keyboard[-1].extend(button)
            return self

        self.keyboard[-1].append(button if isinstance(button, dict) else button.get_data())
        return self

    def row(self) -> typing.Self:
        if len(self.keyboard) and not len(self.keyboard[-1]):
            return self

        self.keyboard.append([])
        return self

    def format_text(self, **format_data: typing.Any) -> typing.Self:
        copy_keyboard = self.copy()

        for row in self.keyboard:
            for button in row:
                button.update(dict(text=button["text"].format(**format_data)))

        return copy_keyboard

    def merge(self, other: BaseKeyboard[KeyboardButton], /) -> typing.Self:
        self.keyboard.extend(copy_keyboard(other.keyboard))
        return self

    def merge_to_last_row(self, other: BaseKeyboard[KeyboardButton], /) -> typing.Self:
        total_rows = len(other.keyboard)

        for index, row in enumerate(copy_keyboard(other.keyboard), start=1):
            for button in row:
                self.keyboard[-1].append(button)

            if index < total_rows:
                self.keyboard.append([])

        return self


__all__ = ("BaseKeyboard",)



import asyncio
import sys
import typing

from telegrinder.api.error import APIServerError, InvalidTokenError
from telegrinder.modules import logger
from telegrinder.tools.aio import maybe_awaitable

if typing.TYPE_CHECKING:
    from telegrinder.bot.polling.polling import Polling


class ErrorHandler:
    _handlers: dict[type[BaseException], typing.Callable[[BaseException], typing.Any]]

    __slots__ = ("_polling", "_handlers")

    def __init__(self, polling: Polling, /) -> None:
        self._polling = polling
        self._handlers = {  # type: ignore
            InvalidTokenError: self._handle_invalid_token_error,
            asyncio.CancelledError: self._handle_cancelled_error,
            APIServerError: self._handle_api_server_error,
            **{e: self._handle_connection_timeout_error for e in polling.api.http.CONNECTION_TIMEOUT_ERRORS},
            **{e: self._handle_client_connection_error for e in polling.api.http.CLIENT_CONNECTION_ERRORS},
        }

    async def handle(self, error: BaseException) -> bool:
        error_class = type(error)

        if error_class is SystemExit:
            self._handle_system_exit(error)  # type: ignore

        if error_class not in self._handlers:
            return False

        try:
            await maybe_awaitable(self._handlers[error_class](error))
            return True
        except SystemExit as sys_exit_err:
            await self._handle_system_exit(sys_exit_err)

    async def _handle_system_exit(self, error: SystemExit) -> typing.NoReturn:
        await logger.aerror("Forced exit from the program with code {}.", error.code)
        raise error from None

    async def _handle_invalid_token_error(
        self,
        error: InvalidTokenError,
    ) -> typing.NoReturn:
        await logger.aerror("{}", error)
        self._polling.stop()
        sys.exit(3)

    async def _handle_cancelled_error(self, _: asyncio.CancelledError) -> None:
        await logger.ainfo("Caught cancel, stopping polling")
        self._polling.stop()

    async def _handle_connection_timeout_error(self, _: BaseException) -> None:
        if self._polling.reconnects_counter > self._polling.max_reconnects:
            await logger.aerror(
                "Failed to reconnect to Telegram API server after {} attempts, stopping polling",
                self._polling.max_reconnects,
            )
            self._polling.stop()
            sys.exit(6)

        await logger.awarning(
            "Server disconnected, waiting {} seconds to reconnect...",
            self._polling.reconnect_after,
        )
        await asyncio.sleep(self._polling.reconnect_after)

    async def _handle_client_connection_error(self, _: BaseException) -> None:
        await logger.aerror(
            "Client connection failed, attempt to reconnect after {} seconds...",
            self._polling.reconnect_after,
        )
        await asyncio.sleep(self._polling.reconnect_after)

    async def _handle_api_server_error(
        self,
        error: APIServerError,
    ) -> None:
        if error.retry_after is None:
            await logger.aerror("{}", error)
            sys.exit(9)

        await logger.aerror("{}, waiting {} seconds to the next request...", error, error.retry_after)
        await asyncio.sleep(error.retry_after)


__all__ = ("ErrorHandler",)



import typing

from kungfu.library.monad.result import Error, Ok, Result
from nodnod.error import NodeError

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import run_post_middleware, run_pre_middleware
from telegrinder.modules import logger
from telegrinder.node.compose import compose
from telegrinder.tools.fullname import fullname
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.handler.abc import ABCHandler
    from telegrinder.bot.dispatch.view.base import View
    from telegrinder.bot.rules.abc import ABCRule


async def process_inner(
    api: API,
    update: Update,
    context: Context,
    view: View,
) -> Result[str, str]:
    for middleware in view.middlewares:
        if await run_pre_middleware(middleware, context) is False:
            await logger.ainfo(
                "Update(id={}, type={!r}) processed with view `{}`. Pre-middleware `{}` raised failure.",
                update.update_id,
                update.update_type,
                view,
                fullname(middleware),
            )
            return Error(f"Pre-middleware `{fullname(middleware)}` raised failure.")

    found_handlers: list[ABCHandler] = []
    responses: list[typing.Any] = []

    for handler in view.handlers:
        match await handler.run(api, update, context):
            case Ok(response):
                found_handlers.append(handler)
                responses.append(response)

                if view.return_manager is not None:
                    await view.return_manager.run(response, api, update, context)

                if handler.final is True:
                    break
            case Error(error):
                await logger.adebug("Running handler `{!r}` failed with error: {}", handler, error)

    context.responses = responses

    for middleware in view.middlewares:
        await run_post_middleware(middleware, context)

    if not found_handlers:
        return Error("No found corresponded handlers.")

    return Ok(
        f"Handler{'s' if len(found_handlers) > 1 else ''}: " + "".join(repr(handler) for handler in found_handlers)
    )


async def check_rule(rule: ABCRule, context: Context) -> bool:
    for requirement in rule.requires:
        if not await check_rule(requirement, context):
            return False

    await logger.adebug("  → Checking rule `{!r}`...", rule)

    async with compose(rule.composable, context) as result:
        match result:
            case Ok(result):
                await logger.adebug("    * Rule `{!r}` is {}", rule, "ok" if result else "failed")
                return result
            case Error(error):
                await logger.adebug(
                    "    * Rule `{}` failed with error:{}\n",
                    fullname(rule),
                    NodeError(f"* failed to compose check of `{fullname(rule)}` rule", from_error=error),
                )

    return False


__all__ = ("check_rule", "process_inner")



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import PaidMediaPurchased, User


class PaidMediaPurchasedCute(BaseCute[PaidMediaPurchased], PaidMediaPurchased, kw_only=True):
    @property
    def from_user(self) -> User:
        return self.from_


__all__ = ("PaidMediaPurchasedCute",)



import asyncio
import typing
from contextlib import suppress
from inspect import isasyncgen, isawaitable

if typing.TYPE_CHECKING:
    from contextvars import Context

type Generator[Yield, Send, Return] = typing.AsyncGenerator[Yield, Send] | typing.Generator[Yield, Send, Return]


def loop_is_running() -> bool:
    with suppress(RuntimeError):
        asyncio.get_running_loop()
        return True

    return False


def get_tasks_results[T](tasks: set[asyncio.Task[T]], /) -> tuple[T, ...]:
    return tuple(task.result() for task in tasks if not task.cancelled() and not task.exception())


def run_task[T](
    task: typing.Awaitable[T],
    /,
    *,
    loop: asyncio.AbstractEventLoop | None = None,
) -> T:
    loop = loop or asyncio.get_event_loop()
    return loop.run_until_complete(future=task)


async def next_generator[T](generator: Generator[T, typing.Any, typing.Any], /) -> T:
    return await send_generator_value(generator, None)


async def send_generator_value[Yield, Send](
    generator: Generator[Yield, Send, typing.Any],
    value: Send | None,
    /,
) -> Yield:
    try:
        return (
            await generator.asend(value) if isasyncgen(generator) else generator.send(value)  # type: ignore
        )
    except StopIteration as exc:
        raise StopGenerator(exc.value, exc.args) from exc


async def stop_generator[Send, Return](
    generator: Generator[typing.Any, Send, Return],
    with_value: Send | None = None,
    /,
) -> Return | None:
    try:
        await send_generator_value(generator, with_value)
    except (StopGenerator, StopAsyncIteration) as exc:
        return exc.value if isinstance(exc, StopGenerator) else None


async def maybe_awaitable[T](obj: T | typing.Awaitable[T], /) -> T:
    if isawaitable(obj):
        return await obj
    return obj


# Source code: https://github.com/facebookincubator/later/blob/main/later/task.py#L68
async def cancel_future(fut: asyncio.Future[typing.Any], /) -> None:
    if fut.done():
        return

    fut.cancel()
    exc: asyncio.CancelledError | None = None

    while not fut.done():
        shielded = asyncio.shield(fut)
        try:
            await asyncio.wait([shielded])
        except asyncio.CancelledError as ex:
            exc = ex
        finally:
            # Insure we handle the exception/value that may exist on the shielded task
            # This will prevent errors logged to the asyncio logger
            if shielded.done() and not shielded.cancelled() and not shielded.exception():
                shielded.result()

    if fut.cancelled():
        if exc is None:
            return
        raise exc from None

    ex = fut.exception()
    if ex is not None:
        raise ex from None

    raise asyncio.InvalidStateError(
        f"Task did not raise CancelledError on cancel: {fut!r} had result {fut.result()!r}",
    )


class StopGenerator(Exception):
    value: typing.Any

    def __init__(self, value: typing.Any, *args: object) -> None:
        super().__init__(*args)
        self.value = value


class TaskGroup[T](asyncio.TaskGroup):
    _all_tasks: set[asyncio.Task[typing.Any]]

    def __init__(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        super().__init__()
        self._all_tasks = set()
        self._loop = loop

    def results(self) -> typing.Iterable[T]:
        while self._all_tasks:
            task = self._all_tasks.pop()
            if not task.cancelled() and task.exception() is None:
                yield task.result()

    def create_task(
        self,
        coro: typing.Coroutine[typing.Any, typing.Any, T],
        /,
        name: str | None = None,
        context: Context | None = None,
    ) -> asyncio.Task[T]:
        task = super().create_task(coro)
        self._all_tasks.add(task)
        return task


__all__ = (
    "StopGenerator",
    "TaskGroup",
    "cancel_future",
    "get_tasks_results",
    "loop_is_running",
    "maybe_awaitable",
    "next_generator",
    "run_task",
    "send_generator_value",
    "stop_generator",
)



import pytest
from nodnod.interface.inject import inject_internals
from nodnod.scope import Scope

from telegrinder.api.api import API
from telegrinder.types.objects import Update


def make_node_scope(api: API, update: Update) -> Scope:
    scope = Scope()
    inject_internals(scope, {API: api, Update: update})
    return scope


@pytest.fixture()
def node_scope_factory(api_instance: API):
    def factory(update: Update) -> Scope:
        return make_node_scope(api_instance, update)

    return factory


@pytest.fixture()
def message_node_scope(api_instance: API, message_update: Update) -> Scope:
    return make_node_scope(api_instance, message_update)


@pytest.fixture()
def callback_query_node_scope(api_instance: API, callback_query_update: Update) -> Scope:
    return make_node_scope(api_instance, callback_query_update)
